/*
 * thread-hangup.cpp
 * Mateusz Szpakowski
 */

//#define USE_NATIVE_THREADS 1

#include <cstdio>
#ifdef USE_NATIVE_THREADS
#  include <windows.h>
#else
#  include <thread>
#endif
#include <FL/Fl.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Window.H>

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

class ThreadApp
{
private:
    Fl_Window* mainWin;
    Fl_Button* startButton;
    Fl_Text_Buffer* msgBuffer;
#ifdef USE_NATIVE_THREADS
    HANDLE workThread;
#else
    std::thread* workThread;
#endif
    unsigned int count;
public:
    ThreadApp();
    ~ThreadApp();
    
    void run();
    static void awakeMsg(void* data);
    static void awakeEnd(void* data);
    static void startCalled(Fl_Widget* w, void* data);
#ifdef USE_NATIVE_THREADS
    static DWORD WINAPI workThreadCallback(LPVOID data);
#endif
    void workThreadFunc();
    static void exiting(Fl_Widget* w, void* data);
};

ThreadApp::ThreadApp()
    : mainWin(nullptr), workThread(nullptr)
{
    mainWin = new Fl_Window(600, 400, "ThreadApp");
    Fl_Text_Display* msgView = new Fl_Text_Display(10, 10, 580, 350);
    msgBuffer = new Fl_Text_Buffer();
    msgView->buffer(msgBuffer);
    startButton = new Fl_Button(10, 360, 580, 30, "START");
    startButton->callback(&ThreadApp::startCalled, this);
    mainWin->callback(&ThreadApp::exiting, this);
    mainWin->end();
}

ThreadApp::~ThreadApp()
{
    if (workThread != nullptr)
#ifdef USE_NATIVE_THREADS
        WaitForSingleObject(workThread, INFINITE);
    CloseHandle(workThread);
#else
        workThread->join();
    delete workThread;
#endif
    delete mainWin;
}

void ThreadApp::run()
{
    Fl::lock();
    mainWin->show();
    Fl::run();
}

void ThreadApp::awakeMsg(void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    char buf[64];
    snprintf(buf, 64, "Message: %u\n", app->count++);
    app->msgBuffer->append(buf);
}

void ThreadApp::awakeEnd(void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->msgBuffer->append("Finish");
    /* finishing thread and freeing its resources */
    if (app->workThread != nullptr)
#ifdef USE_NATIVE_THREADS
        WaitForSingleObject(app->workThread, INFINITE);
    CloseHandle(app->workThread);
#else
        app->workThread->join();
    delete app->workThread;
#endif
    app->workThread = nullptr;
    app->startButton->activate(); // enable
}

void ThreadApp::startCalled(Fl_Widget* w, void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    if (app->workThread != nullptr)
        return; // ignore if ran
    // starting thread
    app->msgBuffer->remove(0, app->msgBuffer->length());
    app->count = 0;
    app->startButton->deactivate();
#ifdef USE_NATIVE_THREADS
    app->workThread = CreateThread(NULL, 32768,
            (LPTHREAD_START_ROUTINE)&ThreadApp::workThreadCallback, app, 0, NULL);
#else
    app->workThread = new std::thread(&ThreadApp::workThreadFunc, app);
#endif
}

#ifdef USE_NATIVE_THREADS
DWORD WINAPI ThreadApp::workThreadCallback(LPVOID data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->workThreadFunc();
    return 0;
}
#endif
void ThreadApp::workThreadFunc()
{
    for (int i = 0; i < 5; i++)
    {
        while (Fl::awake(&ThreadApp::awakeMsg, this) != 0);
#ifdef USE_NATIVE_THREADS
        ::Sleep(1000);
#else
        std::this_thread::sleep_for(std::chrono::milliseconds(1000));
#endif
    }
    // end of thread
    while (Fl::awake(&ThreadApp::awakeEnd, this) != 0);
}

void ThreadApp::exiting(Fl_Widget* w, void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    if (app->workThread == nullptr)
        w->hide();
}

int main(int argc, const char** argv)
{
    ThreadApp app;
    app.run();
    return 0;
}
